"""Project any dataframe, inverse transform and compute stats."""
from typing import Any, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numpy.typing import NDArray

import saiph.reduction.famd as famd
import saiph.reduction.mca as mca
import saiph.reduction.pca as pca
from saiph.models import Model, Parameters


def fit(
    df: pd.DataFrame,
    nf: Optional[Union[int, str]] = None,
    col_w: Optional[NDArray[Any]] = None,
    scale: bool = True,
) -> Tuple[pd.DataFrame, Model, Parameters]:
    """Fit a PCA, MCA or FAMD model on data, imputing what has to be used.

    Parameters
    ----------
    df: pd.DataFrame
        Data to project.
    nf: int|str, default: 'all'
        Number of components to keep.
    col_w: np.ndarrayn default: np.ones(df.shape[1])
        Weight assigned to each variable in the projection
        (more weight = more importance in the axes).
    scale: bool
        Unused. Kept for compatibility with model enabling scale=True|False.

    Returns
    -------
    coord: pd.DataFrame
        The transformed data.
    model: Model
        The model for transforming new data.
    param: Parameters
        The parameters for transforming new data.
    """
    datetime_variables = []
    for i in range(0, df.shape[1]):
        if df.iloc[:, i].dtype == ("datetime64[ns]"):
            df.iloc[:, i] = (
                df.iloc[:, i] - np.datetime64("1970-01-01T00:00:00Z")
            ) / np.timedelta64(1, "s")
            datetime_variables.append(i)

    # Check column types
    quanti = df.select_dtypes(include=["int", "float", "number"]).columns.values
    quali = df.select_dtypes(exclude=["int", "float", "number"]).columns.values

    _nf: int
    if not nf or isinstance(nf, str):
        _nf = min(pd.get_dummies(df).shape)
    else:
        _nf = nf

    # Specify the correct function
    if quali.size == 0:
        _fit = pca.fit
    elif quanti.size == 0:
        _fit = mca.fit
    else:
        _fit = famd.fit

    coord, model, param = _fit(df, _nf, col_w, scale)
    param.quanti = quanti
    param.quali = quali
    param.datetime_variables = np.array(datetime_variables)
    param.cor = _variable_correlation(model, param)

    if quanti.size == 0:
        model.variable_coord = pd.DataFrame(model.D_c @ model.V.T)
    else:
        model.variable_coord = pd.DataFrame(model.V.T)

    return coord, model, param


def stats(model: Model, param: Parameters) -> Parameters:
    """Compute the correlation, contributions and cos2 for each variable."""
    # Check attributes type
    if param.cor is None or param.quanti is None or param.quali is None:
        raise ValueError(
            "empty param, run fit function to create Model class and Parameters class objects"
        )
    model.variable_coord.columns = param.cor.columns
    model.variable_coord.index = list(param.cor.index)

    if param.quali.size == 0:
        param.cos2 = param.cor.applymap(lambda x: x ** 2)
        param.contrib = param.cos2.div(param.cos2.sum(axis=0), axis=1).applymap(
            lambda x: x * 100
        )
    elif param.quanti.size == 0:
        param = mca.stats(model, param)
        if param.cor is None:
            raise ValueError(
                "empty param, run fit function to create Model class and Parameters class objects"
            )
        param.cos2 = param.cor.applymap(lambda x: x ** 2)
        param.contrib = pd.DataFrame(
            param.contrib,
            columns=param.cor.columns,
            index=list(param.cor.index),
        )
    else:
        param = famd.stats(model, param)
        if param.cor is None or param.quanti is None or param.quali is None:
            raise ValueError(
                "empty param, run fit function to create Model class and Parameters class objects"
            )
        param.cos2 = pd.DataFrame(
            param.cos2, index=list(param.quanti) + list(param.quali)
        )
        param.contrib = pd.DataFrame(
            param.contrib,
            columns=param.cor.columns,
            index=list(param.cor.index),
        )
    return param


def transform(df: pd.DataFrame, model: Model, param: Parameters) -> pd.DataFrame:
    """Scale and project into the fitted numerical space.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame to transform.
    model: Model
        Model computed by fit.
    param: Parameters
        Param computed by fit.

    Returns
    -------
    coord: pd.DataFrame
        Coordinates of the dataframe in the fitted space.
    """
    if param.quali is None or param.quanti is None or param.datetime_variables is None:
        raise ValueError("Need to fit before using transform")
    for i in param.datetime_variables:
        df.iloc[:, i] = (
            df.iloc[:, i] - np.datetime64("1970-01-01T00:00:00Z")
        ) / np.timedelta64(1, "s")
    if param.quali.size == 0:
        coord = pca.transform(df, model, param)
    elif param.quanti.size == 0:
        coord = mca.transform(df, model, param)
    else:
        coord = famd.transform(df, model, param)
    return coord


def _variable_correlation(model: Model, param: Parameters) -> pd.DataFrame:
    """Compute the correlation between the axis' and the variables."""
    # select columns and project data
    df_quanti = model.df[param.quanti]
    coord = transform(model.df, model, param)  # transform to be fixed

    if param.quali is not None and len(param.quali) > 0:
        df_quali = pd.get_dummies(model.df[param.quali].astype("category"))
        bind = pd.concat([df_quanti, df_quali], axis=1)
    else:
        bind = df_quanti

    # compute correlations
    cor = pd.DataFrame(
        {
            component: {
                var: coord[component].corr(bind[var].reset_index(drop=True))
                for var in bind.columns
            }
            for component in coord.columns
        }
    )
    return cor


def inverse_transform(
    coord: pd.DataFrame,
    model: Model,
    param: Parameters,
    seed: Optional[int] = None,
) -> pd.DataFrame:
    """Compute the inverse transform of data coordinates.

    Note that if nf was stricly smaller than max(df.shape) in fit,
    inverse_transform o transform != id

    Parameters
    ----------
    coord: pd.DataFrame
        DataFrame to transform.
    model: Model
        Model computed by fit.
    param: Parameters
        Param computed by fit.
    seed: int|None, default: None
        Specify the seed for np.random.

    Returns
    -------
    inverse: pd.DataFrame
        Inversed DataFrame.
    """
    np.random.seed(seed)

    if len(coord) < param.nf:
        raise ValueError(
            "Inverse_transform is not working"
            "if the number of dimensions is greater than the number of individuals"
        )

    # if PCA or FAMD compute the continuous variables
    if param.quanti is not None and len(param.quanti) != 0:

        X = np.array(coord @ model.V * np.sqrt(param.col_w))
        X = X / np.sqrt(param.col_w) * param.col_w
        X_quanti = X[:, : len(param.quanti)]

        # descale
        std = np.array(model.std)
        mean = np.array(model.mean)
        inverse_quanti = (X_quanti * std) + mean
        inverse_quanti = pd.DataFrame(inverse_quanti, columns=list(param.quanti))

        # if FAMD descale the categorical variables
        if param.quali is not None and len(param.quali) != 0:
            X_quali = X[:, len(param.quanti) :]
            prop = np.array(model.prop)
            X_quali = (X_quali) * (np.sqrt(prop)) + prop

    # if MCA no descaling
    else:
        # Previously was a ndarray, but no need
        # NB: If this causes a bug, X_quali = np.array(X_quali) goes back to previous vesrion
        X_quali = coord @ (model.D_c @ model.V.T).T  # type: ignore
        X_quali = np.divide(X_quali, param.dummies_col_prop)
        # divinding by proportion of each modality among individual
        # allows to get back the complete disjunctive table
        # X_quali is the complete disjunctive table ("tableau disjonctif complet" in FR)

    # compute the categorical variables
    if param.quali is not None and len(param.quali) != 0:
        inverse_quali = pd.DataFrame()
        X_quali = pd.DataFrame(X_quali)
        X_quali.columns = list(
            pd.get_dummies(
                model.df[param.quali],
                prefix=None,
                prefix_sep="_",
            ).columns
        )

        modalities = []
        for column in param.quali:
            modalities += [len(model.df[column].unique())]
        val = 0
        # conserve the modalities in their original type
        modalities_type = []
        for col in param.quali:
            mod_temp = list(model.df[col].unique())
            mod_temp.sort()  # sort the modalities as pd.get_dummies have done
            modalities_type += mod_temp

        # create a dict that link dummies variable to the original modalitie
        dict_mod = dict(zip(X_quali.columns, modalities_type))

        # for each variable we affect the value to the highest modalitie in X_quali
        for i in range(len(modalities)):
            # get cumululative probabilities
            c = X_quali.iloc[:, val : val + modalities[i]].cumsum(axis=1)
            # random draw
            u = np.random.rand(len(c), 1)
            # choose the modality according the probabilities of each modalities
            mod_random = (u < c).idxmax(axis=1)
            mod_random = [dict_mod.get(x, x) for x in mod_random]
            inverse_quali[list(model.df[param.quali].columns)[i]] = mod_random
            val += modalities[i]

    # concatenate the continuous and categorical
    if (
        param.quali is not None
        and param.quanti is not None
        and len(param.quali) != 0
        and len(param.quanti) != 0
    ):
        inverse = pd.concat([inverse_quali, inverse_quanti], axis=1)
    elif param.quanti is not None and len(param.quanti) != 0:
        inverse = inverse_quanti
    else:
        inverse = inverse_quali

    # Cast columns to same type as input
    for column in model.df.columns:
        column_type = model.df.loc[:, column].dtype
        inverse[column] = inverse[column].astype(column_type)

    # reorder back columns
    inverse = inverse[model.df.columns]

    # Turn back datetime variables to original dtype
    if param.datetime_variables is not None:
        for i in param.datetime_variables:
            inverse.iloc[:, i] = (
                inverse.iloc[:, i] * np.timedelta64(1, "s")
            ) + np.datetime64("1970-01-01T00:00:00Z")

    return inverse


def decimal_count(n: float) -> int:
    """Return the decimal part of the number."""
    f = str(n)
    if "." in f:
        digits = f[::-1].find(".")
    else:
        digits = 0
    return digits



import pandas as pd
import pytest

_iris_csv = pd.read_csv("fixtures/iris.csv")


@pytest.fixture
def iris_df() -> pd.DataFrame:
    return _iris_csv


@pytest.fixture
def iris_quanti_df() -> pd.DataFrame:
    return _iris_csv.drop("variety", axis=1)



"""PCA projection."""
import sys
from typing import Any, Optional, Tuple

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from saiph.models import Model, Parameters
from saiph.reduction.utils.check_params import fit_check_params
from saiph.reduction.utils.common import (
    column_names,
    explain_variance,
    row_weights_uniform,
)
from saiph.reduction.utils.svd import SVD


def fit(
    df: pd.DataFrame,
    nf: Optional[int] = None,
    col_w: Optional[NDArray[Any]] = None,
    scale: Optional[bool] = True,
) -> Tuple[pd.DataFrame, Model, Parameters]:
    """Fit a PCA model on data.

    Parameters
    ----------
    df: pd.DataFrame
        Data to project.
    nf: int, default: min(df.shape)
        Number of components to keep.
    col_w: np.ndarrayn default: np.ones(df.shape[1])
        Weight assigned to each variable in the projection
        (more weight = more importance in the axes).
    scale: bool
        Wether to scale data or not.

    Returns
    -------
    coord: pd.DataFrame
        The transformed data.
    model: Model
        The model for transforming new data.
    param: Parameters
        The parameters for transforming new data.
    """
    nf = nf or min(df.shape)
    _col_weights = col_w or np.ones(df.shape[1])
    if not isinstance(df, pd.DataFrame):
        df = pd.DataFrame(df)
    fit_check_params(nf, _col_weights, df.shape[1])

    # set row weights
    row_w = row_weights_uniform(len(df))

    df_centered, mean, std = center(df, scale)

    # apply weights and compute svd
    Z = ((df_centered * _col_weights).T * row_w).T
    U, s, V = SVD(Z)

    U = ((U.T) / np.sqrt(row_w)).T
    V = V / np.sqrt(_col_weights)

    explained_var, explained_var_ratio = explain_variance(s, df_centered, nf)

    U = U[:, :nf]
    s = s[:nf]
    V = V[:nf, :]

    columns = column_names(nf)
    coord = df_centered @ V.T
    coord.columns = columns

    model = Model(
        df=df,
        U=U,
        V=V,
        explained_var=explained_var,
        explained_var_ratio=explained_var_ratio,
        variable_coord=pd.DataFrame(V.T),
        mean=mean,
        std=std,
        type="pca",
    )

    param = Parameters(nf=nf, col_w=_col_weights, row_w=row_w, columns=columns)

    return coord, model, param


def center(
    df: pd.DataFrame, scale: Optional[bool] = True
) -> Tuple[pd.DataFrame, float, float]:
    """Center data and standardize it if scale == true.

    Used as internal function during fit. Scaler is better suited when a Model is already fitted.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame to center.
    scale: bool
        Whether dataframe should be standardized ot not.

    Returns
    -------
    df: pd.DataFrame
        The centered DataFrame.
    mean: float
        Mean of the input dataframe.
    std: float
        Standard deviation of the input dataframe. Returns nan as std if no std was asked.
    """
    df = df.copy()
    mean = np.mean(df, axis=0)
    df -= mean
    std = 0
    if scale:
        std = np.std(df, axis=0)
        std[std <= sys.float_info.min] = 1  # type: ignore
        df /= std
    # else:
    #     std = np.nan
    return df, mean, std


def scaler(model: Model, df: Optional[pd.DataFrame] = None) -> pd.DataFrame:
    """Scale data using mean and std from model.

    Parameters
    ----------
    model: Model
        Model computed by fit.
    df: pd.DataFrame
        DataFrame to scale.
        If nothing is specified, takes the DataFrame on which the model was fitted.

    Returns
    -------
    df: pd.DataFrame
        The scaled DataFrame.
    """
    if df is None:
        df = model.df

    df_scaled = df.copy()

    df_scaled -= model.mean
    df_scaled /= model.std
    return df_scaled


def transform(df: pd.DataFrame, model: Model, param: Parameters) -> pd.DataFrame:
    """Scale and project into the fitted numerical space.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame to transform.
    model: Model
        Model computed by fit.
    param: Parameters
        Param computed by fit.

    Returns
    -------
    coord: pd.DataFrame
        Coordinates of the dataframe in the fitted space.
    """
    df_scaled = scaler(model, df)
    coord = df_scaled @ model.V.T
    coord.columns = param.columns
    return coord



from dataclasses import dataclass
from typing import Any, List, Optional

import pandas as pd
from numpy.typing import NDArray


@dataclass
class Model:
    df: pd.DataFrame
    """DataFrame on which the model was fit."""

    explained_var: NDArray[Any]
    """Explained variance."""
    explained_var_ratio: NDArray[Any]
    """Explained variance divided by the sum of the variances."""
    variable_coord: pd.DataFrame

    V: NDArray[Any]
    """Orthogonal matrix with right singular vectors as rows."""
    U: Optional[NDArray[Any]] = None
    """Orthogonal matrix with left singular vectors as columns."""
    s: Optional[NDArray[Any]] = None
    """Singular values"""

    mean: Optional[float] = None
    """Mean of the original data. Calculated while centering."""
    std: Optional[float] = None
    """Standard deviation of the original data. Calculated while scaling."""
    prop: Any = None  # FAMD only
    """Modality proportions of categorical variables"""
    _modalities: Optional[NDArray[Any]] = None
    D_c: Optional[NDArray[Any]] = None

    type: Optional[str] = None
    """Type of dimension reduction that was performed."""


@dataclass
class Parameters:
    nf: int
    """Number of components kept."""
    col_w: NDArray[Any]
    """Weights that were applied to each column."""
    row_w: NDArray[Any]
    """Weights that were applied to each row."""
    columns: List[Any]
    quanti: Optional[NDArray[Any]] = None
    """Indices of columns that are considered quantitative."""
    quali: Optional[NDArray[Any]] = None
    """Indices of columns that are considered qualitative."""
    datetime_variables: Optional[NDArray[Any]] = None
    """Indices of columns that are considered datetimes."""
    cor: Optional[pd.DataFrame] = None
    contrib: Optional[pd.DataFrame] = None
    cos2: Optional[pd.DataFrame] = None
    dummies_col_prop: Optional[NDArray[Any]] = None
    """Proportion of individuals taking each modality."""



from itertools import repeat
from typing import Any, List, Tuple

import numpy as np
import pandas as pd
import scipy
from numpy.typing import NDArray


def column_names(n: int) -> List[str]:
    return [f"Dim. {i + 1}" for i in range(n)]


def row_weights_uniform(n: int) -> NDArray[np.float64]:
    return np.array([k for k in repeat(1 / n, n)])


def diag(arr: NDArray[Any], use_scipy: bool = False) -> NDArray[Any]:
    if use_scipy:
        return scipy.sparse.diags(arr)  # type: ignore
    else:
        return np.diag(arr)  # type: ignore


def explain_variance(
    s: NDArray[Any], df: pd.DataFrame, nf: int
) -> Tuple[NDArray[Any], NDArray[Any]]:
    explained_var: NDArray[Any] = ((s ** 2) / (df.shape[0] - 1))[:nf]
    summed_explained_var = explained_var.sum()
    if summed_explained_var == 0:
        explained_var_ratio = np.array([np.nan])
    else:
        explained_var_ratio = explained_var / explained_var.sum()
    return explained_var, explained_var_ratio



import numpy as np
import pandas as pd
import prince
from numpy.testing import assert_allclose
from pandas.testing import assert_frame_equal
from sklearn import decomposition

from saiph.reduction.pca import center, fit, scaler, transform

# mypy: ignore-errors


def test_fit_scale() -> None:
    scale = True

    df = pd.DataFrame(
        {
            "one": [1.0, 3.0],
            "two": [2.0, 4.0],
        }
    )

    result, model, _ = fit(df, scale=scale)

    expected_result = pd.DataFrame(
        {
            "Dim. 1": [-1.41, 1.41],
            "Dim. 2": [0.0, 0.0],
        }
    )
    expected_v = np.array([[0.71, 0.71], [-0.71, 0.71]])
    expected_explained_var = np.array([1.0, 0.0])
    expected_explained_var_ratio = np.array([1.0, 0.0])

    assert_frame_equal(result, expected_result, check_exact=False, atol=0.01)

    assert_frame_equal(model.df, df)
    assert_allclose(model.V, expected_v, atol=0.01)
    assert_allclose(model.explained_var, expected_explained_var, atol=0.01)
    assert_allclose(model.explained_var_ratio, expected_explained_var_ratio, atol=0.01)
    assert_allclose(model.variable_coord, model.V.T, atol=0.01)
    assert_allclose(model.mean, [2.0, 3.0])
    assert_allclose(model.std, [1.0, 1.0])


def test_fit_not_scale() -> None:
    scale = False

    df = pd.DataFrame(
        {
            "one": [1.0, 2.0, 3.0],
            "two": [2.0, 4.0, 5.0],
        }
    )

    result, model, _ = fit(df, scale=scale)

    expected_result = pd.DataFrame(
        {
            "Dim. 1": [1.94, -0.28, -1.66],
            "Dim. 2": [-0.07, 0.18, -0.11],
        }
    )
    expected_v = np.array([[-0.544914, -0.838492], [-0.838492, 0.544914]])
    expected_explained_var = np.array([0.367571, 0.002799])
    expected_explained_var_ratio = np.array([1.0, 0.0])

    assert_frame_equal(result, expected_result, check_exact=False, atol=0.01)

    assert_frame_equal(model.df, df)
    assert_allclose(model.V, expected_v, atol=0.01)
    assert_allclose(model.explained_var, expected_explained_var, atol=0.01)
    assert_allclose(model.explained_var_ratio, expected_explained_var_ratio, atol=0.01)
    assert_allclose(model.variable_coord, model.V.T, atol=0.01)  # type: ignore
    assert_allclose(model.mean, [2.0, 3.666667])
    # Default value when not scaling
    assert_allclose(model.std, 0.0)


def test_fit_zero() -> None:
    scale = True

    df = pd.DataFrame(
        {
            "one": [1, 1],
            "two": [2, 2],
        }
    )

    result, model, _ = fit(df, scale=scale)

    expected_result = pd.DataFrame(
        {
            "Dim. 1": [0.0, 0.0],
            "Dim. 2": [0.0, 0.0],
        }
    )
    expected_v = np.array([[1.0, 0.0], [0.0, 1.0]])
    expected_explained_var = np.array([0.0, 0.0])

    assert_frame_equal(result, expected_result, check_exact=False, atol=0.01)

    assert_frame_equal(model.df, df)
    assert_allclose(model.V, expected_v, atol=0.01)
    assert_allclose(model.explained_var, expected_explained_var, atol=0.01)
    # np.nan == np.nan returns False
    assert pd.isnull(model.explained_var_ratio)
    assert_allclose(model.variable_coord, model.V.T, atol=0.01)
    assert_allclose(model.mean, [1.0, 2.0])
    assert_allclose(model.std, [1.0, 1.0])


def test_center_scaler() -> None:
    df = pd.DataFrame(
        {
            0: [1.0, 12.0],
            1: [2.0, 4.0],
        }
    )

    _, model, _ = fit(df, scale=True)

    df1, _, _ = center(model.df, scale=True)
    df2 = scaler(model, None)

    assert_frame_equal(df1, df2)


def test_transform_simple() -> None:
    df = pd.DataFrame(
        {
            0: [1.0, 12.0],
            1: [2.0, 4.0],
        }
    )

    _, model, param = fit(df, scale=True)

    df_transformed = transform(df, model, param)

    expected_transformed = pd.DataFrame(
        {
            "Dim. 1": [-1.414214, 1.414214],
            "Dim. 2": [0.0, 0.0],
        }
    )

    assert_frame_equal(df_transformed, expected_transformed, atol=0.01)


def test_transform() -> None:
    df = pd.DataFrame({0: [-2.0, 7.0, -4.5], 1: [6.0, 2.0, 7.0], 2: [5.0, 10.0, -14.5]})

    _, model, param = fit(df, scale=True)

    df_transformed = transform(df, model, param)

    expected_transformed = pd.DataFrame(
        {
            "Dim. 1": [-0.285647, 2.150812, -1.865165],
            "Dim. 2": [0.730941, -0.287485, -0.443456],
            "Dim. 3": [0.0, 0.0, 0.0],
        }
    )

    assert_frame_equal(
        df_transformed, expected_transformed, check_exact=False, atol=0.00001
    )


def test_transform_vs_coord() -> None:
    df = pd.DataFrame({0: [-2.0, 7.0, -4.5], 1: [6.0, 2.0, 7.0], 2: [5.0, 10.0, -14.5]})
    coord, model, param = fit(df, scale=True)
    df_transformed = transform(df, model, param)

    assert_frame_equal(coord, df_transformed)


def test_compare_sklearn_simple() -> None:
    df = pd.DataFrame(
        {
            0: [1.0, 12.0],
            1: [2.0, 4.0],
        }
    )

    coord, _, _ = fit(df, scale=False)
    pca = decomposition.PCA(n_components=2)
    coord_scikit = pca.fit_transform(df)

    assert_allclose(coord.to_numpy(), coord_scikit, atol=0.0001)


def test_compare_sklearn_full(iris_quanti_df) -> None:
    coord, _, _ = fit(iris_quanti_df, scale=False)
    pca = decomposition.PCA(n_components=4)
    coord_scikit = pca.fit_transform(iris_quanti_df)

    assert_allclose(coord.to_numpy(), coord_scikit, atol=0.0001)


def test_compare_sklearn_nf_reduced(iris_quanti_df) -> None:
    coord, _, _ = fit(iris_quanti_df, scale=False, nf=3)
    pca = decomposition.PCA(n_components=3)
    coord_scikit = pca.fit_transform(iris_quanti_df)

    assert_allclose(coord.to_numpy(), coord_scikit, atol=0.0001)


def test_compare_prince_full(iris_quanti_df) -> None:
    pca = prince.PCA(n_components=4, rescale_with_std=False)
    pca = pca.fit(iris_quanti_df)
    pca = pca.transform(iris_quanti_df)

    coord, _, _ = fit(iris_quanti_df, scale=False)

    assert_allclose(coord.to_numpy(), pca, atol=0.0001)



import numpy as np
import pandas as pd
from numpy.testing import assert_allclose
from pandas.testing import assert_frame_equal

from saiph.reduction.mca import fit, transform

# mypy: ignore-errors


def test_fit() -> None:
    df = pd.DataFrame(
        {
            "tool": ["toaster", "hammer"],
            "score": ["aa", "aa"],
        }
    )

    result, model, _ = fit(df)

    expected_result = pd.DataFrame(
        {
            "Dim. 1": [0.7, -0.7],
            "Dim. 2": [-0.7, -0.7],
        }
    )
    expected_v = np.array([[-0.707107, 0.707107, -0.0], [-0.707107, -0.707107, 0.0]])
    expected_explained_var = np.array([1.25000e-01, 3.85186e-34])
    expected_explained_var_ratio = np.array([1.0, 0.0])

    assert_frame_equal(result, expected_result, check_exact=False, atol=0.01)

    assert_frame_equal(model.df, df)
    assert_allclose(model.V, expected_v, atol=0.01)
    assert_allclose(model.explained_var, expected_explained_var, atol=0.01)
    assert_allclose(model.explained_var_ratio, expected_explained_var_ratio, atol=0.01),
    assert np.array_equal(
        model._modalities, ["tool_hammer", "tool_toaster", "score_aa"]
    )
    assert_allclose(
        model.D_c,
        np.array([[2.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 1.41421356]]),
        atol=0.01,
    )
    assert model.mean is None
    assert model.std is None


def test_fit_zero() -> None:
    df = pd.DataFrame(
        {
            "tool": ["toaster", "toaster"],
            "score": ["aa", "aa"],
        }
    )

    result, model, _ = fit(df)

    expected_result = pd.DataFrame(
        {
            "Dim. 1": [0.7, 0.7],
            "Dim. 2": [0.7, 0.7],
        }
    )
    expected_v = np.array([[1.0, 0.0], [0.0, 1.0]])
    expected_explained_var = np.array([0.0, 0.0])

    assert_frame_equal(result, expected_result, check_exact=False, atol=0.01)
    assert_frame_equal(model.df, df)
    assert_allclose(model.V, expected_v, atol=0.01)
    assert_allclose(model.explained_var, expected_explained_var, atol=0.01)
    assert pd.isna(model.explained_var_ratio)
    assert np.array_equal(model._modalities, ["tool_toaster", "score_aa"])
    assert_allclose(
        model.D_c,
        np.array([[1.414214, 0.0], [0.0, 1.414214]]),
        atol=0.01,
    )
    assert model.mean is None
    assert model.std is None


def test_fit_zero_same_df() -> None:
    """Verify that we get the same result if the pattern matches."""
    df = pd.DataFrame(
        {
            "tool": ["toaster", "toaster"],
            "score": ["aa", "aa"],
        }
    )
    df_2 = pd.DataFrame(
        {
            "tool": ["hammer", "hammer"],
            "score": ["bb", "bb"],
        }
    )

    result1, model1, _ = fit(df)
    result2, model2, _ = fit(df_2)

    assert_frame_equal(result1, result2)

    for k in [
        "explained_var",
        "variable_coord",
        "variable_coord",
        "U",
        "s",
        "mean",
        "std",
        "prop",
        "D_c",
    ]:  # removed "_modalities", "df", "explained_var_ratio"
        k1 = getattr(model1, k)
        k2 = getattr(model2, k)
        if isinstance(k1, pd.DataFrame):
            assert k1.equals(k2)
        elif isinstance(k1, np.ndarray):
            assert np.array_equal(k1, k2)
        else:
            assert k1 == k2


def test_transform_simple() -> None:
    df = pd.DataFrame(
        {
            "tool": ["toaster", "toaster"],
            "score": ["aa", "aa"],
        }
    )
    _, model, param = fit(df, scale=True)

    df_transformed = transform(df, model, param)

    print(df_transformed)

    expected_transform = pd.DataFrame(
        {
            "Dim. 1": [0.707107, 0.707107],
            "Dim. 2": [0.707107, 0.707107],
        }
    )

    assert_frame_equal(
        df_transformed, expected_transform, check_exact=False, atol=0.00001
    )


def test_transform_vs_coord() -> None:
    df = pd.DataFrame(
        {
            "tool": ["toaster", "toaster"],
            "score": ["aa", "aa"],
        }
    )
    coord, model, param = fit(df)
    df_transformed = transform(df, model, param)

    assert_frame_equal(coord, df_transformed)



import pandas as pd
import pytest
from numpy.testing import assert_allclose
from pandas.testing import assert_frame_equal

from saiph import fit, inverse_transform, stats, transform

# mypy: ignore-errors


def test_transform_then_inverse_FAMD(iris_df: pd.DataFrame) -> None:
    _, model, param = fit(iris_df, nf="all")
    transformed = transform(iris_df, model, param)
    un_transformed = inverse_transform(transformed, model, param)
    print(un_transformed)
    print(iris_df)

    assert_frame_equal(un_transformed, iris_df)


def test_transform_then_inverse_PCA(iris_quanti_df: pd.DataFrame) -> None:
    _, model, param = fit(iris_quanti_df, nf="all")
    transformed = transform(iris_quanti_df, model, param)
    un_transformed = inverse_transform(transformed, model, param)
    assert_frame_equal(un_transformed, iris_quanti_df)


def test_transform_then_inverse_MCA() -> None:
    df = pd.DataFrame(
        {
            "tool": [
                "toaster",
                "toaster",
                "hammer",
                "toaster",
                "toaster",
                "hammer",
                "toaster",
                "toaster",
                "hammer",
            ],
            "score": ["aa", "ca", "bb", "aa", "ca", "bb", "aa", "ca", "bb"],
            "car": [
                "tesla",
                "renault",
                "tesla",
                "tesla",
                "renault",
                "tesla",
                "tesla",
                "renault",
                "tesla",
            ],
            "moto": [
                "Bike",
                "Bike",
                "Motor",
                "Bike",
                "Bike",
                "Motor",
                "Bike",
                "Bike",
                "Motor",
            ],
        }
    )

    _, model, param = fit(df)
    transformed = transform(df, model, param)
    un_transformed = inverse_transform(transformed, model, param)
    assert_frame_equal(un_transformed, df)


def test_transform_then_inverse_MCA_type() -> None:
    df = pd.DataFrame(
        {
            "tool": [
                "toaster",
                "toaster",
                "hammer",
                "toaster",
                "toaster",
                "hammer",
                "toaster",
                "toaster",
                "hammer",
            ],
            "score": [1, 1, 0, 1, 1, 1, 1, 0, 0],
            "car": [
                "tesla",
                "renault",
                "tesla",
                "tesla",
                "renault",
                "tesla",
                "tesla",
                "renault",
                "tesla",
            ],
            "moto": [
                "Bike",
                "Bike",
                "Motor",
                "Bike",
                "Bike",
                "Motor",
                "Bike",
                "Bike",
                "Motor",
            ],
        }
    )

    df = df.astype("object")
    _, model, param = fit(df)
    transformed = transform(df, model, param)
    un_transformed = inverse_transform(transformed, model, param)

    assert_frame_equal(un_transformed, df)


def test_transform_then_inverse_FAMD_weighted() -> None:
    df = pd.DataFrame(
        {
            "variable_1": [4, 5, 6, 7, 11, 2, 52],
            "variable_2": [10, 20, 30, 40, 10, 74, 10],
            "variable_3": ["red", "blue", "blue", "green", "red", "blue", "red"],
            "variable_4": [100, 50, -30, -50, -19, -29, -20],
        }
    )

    _, model, param = fit(df, col_w=[2, 1, 3, 2])
    transformed = transform(df, model, param)
    un_transformed = inverse_transform(transformed, model, param)

    assert_frame_equal(un_transformed, df)


def test_transform_then_inverse_PCA_weighted() -> None:
    df = pd.DataFrame(
        {
            "variable_1": [4, 5, 6, 7, 11, 2, 52],
            "variable_2": [10, 20, 30, 40, 10, 74, 10],
            "variable_3": [100, 50, -30, -50, -19, -29, -20],
        }
    )

    _, model, param = fit(df, col_w=[2, 1, 3])
    transformed = transform(df, model, param)
    un_transformed = inverse_transform(transformed, model, param)

    assert_frame_equal(un_transformed, df)


def test_transform_then_inverse_MCA_weighted() -> None:
    df = pd.DataFrame(
        {
            "variable_1": ["1", "3", "3", "3", "1", "2", "2", "1", "1", "2"],
            "variable_2": ["1", "1", "1", "2", "2", "1", "1", "1", "2", "2"],
            "variable_3": ["1", "2", "1", "2", "1", "2", "1", "1", "2", "2"],
            "variable_4": [
                "red",
                "blue",
                "blue",
                "green",
                "red",
                "blue",
                "red",
                "red",
                "red",
                "red",
            ],
        }
    )

    _, model, param = fit(df, col_w=[2, 1, 3, 2])
    transformed = transform(df, model, param)
    un_transformed = inverse_transform(transformed, model, param)

    assert_frame_equal(un_transformed, df)


def test_coords_vs_transform_with_multiple_nf(iris_df: pd.DataFrame) -> None:
    with pytest.raises(ValueError):
        fit(iris_df, nf=10000)

    with pytest.raises(ValueError):
        fit(iris_df, nf=-1)

    for n in range(7):
        coord, model, param = fit(iris_df, nf=n)
        transformed = transform(iris_df, model, param)
        assert_frame_equal(coord, transformed)


df_pca = pd.DataFrame(
    {
        0: [
            1000.0,
            3000.0,
            10000.0,
            1500.0,
            700.0,
            3300.0,
            5000.0,
            2000.0,
            1200.0,
            6000.0,
        ],
        1: [185.0, 174.3, 156.8, 182.7, 180.3, 179.2, 164.7, 192.5, 191.0, 169.2],
        2: [1, 5, 10, 2, 4, 4, 7, 3, 1, 6],
    }
)

df_famd = pd.DataFrame(
    {
        "variable_1": [4, 5, 6, 7, 11, 2, 52],
        "variable_2": [10, 20, 30, 40, 10, 74, 10],
        "variable_3": ["red", "blue", "blue", "green", "red", "blue", "red"],
        "variable_4": [100, 50, -30, -50, -19, -29, -20],
    }
)

df_mca = pd.DataFrame(
    {
        0: [
            "red",
            "red",
            "whithe",
            "whithe",
            "red",
            "whithe",
            "red",
            "red",
            "whithe",
            "red",
        ],
        1: [
            "beef",
            "chicken",
            "fish",
            "fish",
            "beef",
            "chicken",
            "beef",
            "chicken",
            "fish",
            "beef",
        ],
        2: [
            "france",
            "espagne",
            "france",
            "italy",
            "espagne",
            "france",
            "france",
            "espagne",
            "chine",
            "france",
        ],
    }
)


@pytest.mark.parametrize(
    "df_input,expected_type", [(df_pca, "pca"), (df_mca, "mca"), (df_famd, "famd")]
)
def test_eval(df_input, expected_type):
    _, model, _ = fit(df_input)
    assert model.type == expected_type


# check contribution of variables to dim

expected_pca_contrib = [32.81178064277649, 33.12227570926467, 34.065943647958846]
expected_mca_contrib = [
    13.314231201732547,
    19.971346802598834,
    7.924987451762375,
    2.8647115861394203,
    24.435070805047193,
    11.119305005676665,
    8.778349565191498,
    0.47269257617482885,
    11.119305005676662,
]
expected_famd_contrib = [
    15.696161557629662,
    36.08406414786589,
    11.420291290785196,
    9.852860955848701,
    6.104123324745425,
    20.842498723125182,
]


@pytest.mark.parametrize(
    "df_input,expected_contrib",
    [
        (df_pca, expected_pca_contrib),
        (df_mca, expected_mca_contrib),
        (df_famd, expected_famd_contrib),
    ],
)
def test_var_contrib(df_input, expected_contrib):
    _, model, param = fit(df_input)
    stats(model, param)
    assert_allclose(param.contrib["Dim. 1"], expected_contrib, atol=1e-07)


# check cor of variables to dim (if cor is ok so is cos2)

expected_pca_cor = [0.9621683005738202, -0.9667100394109722, 0.9803843201246043]
expected_mca_cor = [
    -0.9169410964961192,
    0.9169410964961192,
    -0.5776131247092218,
    -0.3215176498417082,
    0.9390120540605189,
    0.5586386162833192,
    -0.5628216589197227,
    -0.15453176858793358,
    0.5586386162833191,
]
expected_famd_cor = [
    -0.5930925452224494,
    0.8992562362632682,
    -0.5058995881660323,
    0.6216213705726737,
    0.399494477072544,
    -0.9041066243572436,
]


@pytest.mark.parametrize(
    "df_input,expected_cor",
    [
        (df_pca, expected_pca_cor),
        (df_mca, expected_mca_cor),
        (df_famd, expected_famd_cor),
    ],
)
def test_var_cor(df_input, expected_cor):
    _, model, param = fit(df_input)
    stats(model, param)
    assert_allclose(param.cor["Dim. 1"], expected_cor, atol=1e-07)


# Check percentage of explained variance

expected_pca_explained_var_ratio = [
    0.9404831846910865,
    0.040151017139633684,
    0.01936579816927985,
]
expected_mca_explained_var_ratio = [
    0.42099362789799255,
    0.23253662367291794,
    0.1666666666666665,
    0.1314458557658021,
    0.035220810900864555,
]
expected_famd_explained_var_ratio = [
    0.44820992177856206,
    0.2801754650846534,
    0.1707856006031922,
    0.05566366403780453,
    0.04516534849578783,
]


@pytest.mark.parametrize(
    "df_input,expected_var_ratio",
    [
        (df_pca, expected_pca_explained_var_ratio),
        (df_mca, expected_mca_explained_var_ratio),
        (df_famd, expected_famd_explained_var_ratio),
    ],
)
def test_var_ratio(df_input, expected_var_ratio):
    _, model, param = fit(df_input)
    stats(model, param)
    assert_allclose(model.explained_var_ratio[0:5], expected_var_ratio, atol=1e-07)



"""Visualization functions."""
from typing import Any, List, Optional

import matplotlib.pyplot as plt  # type: ignore
import numpy as np
import pandas as pd
from numpy.typing import NDArray

from saiph.models import Model, Parameters


def plot_circle(
    model: Model,
    param: Parameters,
    dimensions: Optional[List[int]] = None,
    min_cor: float = 0.1,
    max_var: int = 7,
) -> None:
    """Plot correlation graph.

    Arguments:
        dimensions: dimensions to help by each axis

    Returns:
        plot of the correlation circle
    """
    # make sure stats have been computed prior to visualization
    if param.cor is None:
        raise ValueError(
            "empty param, run fit function to create Model class and Parameters class objects"
        )

    # Dimensions start from 1

    # Plotting circle
    dimensions = dimensions or [1, 2]
    figure_axis_size = 6
    explained_var_ratio = model.explained_var_ratio

    circle1 = plt.Circle((0, 0), radius=1, color="k", fill=False)
    fig = plt.gcf()
    fig.set_size_inches(5, 5)
    fig.gca().add_artist(circle1)

    # Order dataframe
    cor = param.cor.copy()
    cor["sum"] = cor.apply(
        lambda x: abs(x[dimensions[0] - 1]) + abs(x[dimensions[1] - 1]), axis=1
    )
    cor.sort_values(by="sum", ascending=False, inplace=True)

    # Plotting arrows
    texts = []
    i = 0
    for name, row in cor.iterrows():
        if i < max_var and (
            np.abs(row[dimensions[0] - 1]) > min_cor
            or np.abs(row[dimensions[1] - 1]) > min_cor
        ):
            x = row[dimensions[0] - 1]
            y = row[dimensions[1] - 1]
            plt.arrow(
                0.0,
                0.0,
                x,
                y,
                color="k",
                length_includes_head=True,
                head_width=0.05,
            )

            plt.plot([0.0, x], [0.0, y], "k-")
            texts.append(plt.text(x, y, name, fontsize=2 * figure_axis_size))
            i += 1

    # Plotting vertical lines
    plt.plot([-1.1, 1.1], [0, 0], "k--")
    plt.plot([0, 0], [-1.1, 1.1], "k--")

    # Setting limits and title
    plt.xlim((-1.1, 1.1))
    plt.ylim((-1.1, 1.1))
    plt.title("Correlation Circle", fontsize=figure_axis_size * 3)

    plt.xlabel(
        "Dim "
        + str(dimensions[0])
        + " (%s%%)" % str(explained_var_ratio[dimensions[0] - 1] * 100)[:4],
        fontsize=figure_axis_size * 2,
    )
    plt.ylabel(
        "Dim "
        + str(dimensions[1])
        + " (%s%%)" % str(explained_var_ratio[dimensions[1] - 1] * 100)[:4],
        fontsize=figure_axis_size * 2,
    )


def plot_var_contribution(
    model: Model,
    param: Parameters,
    dim: int = 1,
    max_var: int = 10,
    min_contrib: float = 0.1,
) -> None:
    """Plot the variable contribution for a given dimension.

    Args:
        dim: value of the dimension to plot
        max_var: maximum number of variables to plot
        min_contrib: lower threshold for the variable contributions

    Returns:
        graph of the contribution percentages per variables
    """
    if param.cos2 is None or param.contrib is None:
        raise ValueError(
            "empty param, run fit function to create Model class and Parameters class objects"
        )
    # Dimensions start from 1

    # get the useful contributions
    var_contrib = param.contrib[param.contrib.columns[dim - 1]]
    if len(var_contrib) > max_var:
        var_contrib = var_contrib[:max_var]

    # check threshold
    var_contrib = [var for var in var_contrib if var > min_contrib]
    var_contrib = pd.DataFrame(var_contrib)[0]

    indices = list((-var_contrib).argsort())
    names = [list(param.contrib.index)[indices[i]] for i in range(len(indices))]

    # plot
    plt.figure(figsize=(12, 6))
    plt.bar(range(len(var_contrib)), var_contrib[indices], align="center")
    plt.xticks(range(len(var_contrib)), names, rotation="horizontal")

    # setting labels and title
    plt.title("Variables contributions to Dim. " + str(dim))
    plt.ylabel("Importance")
    plt.xlabel("Variables")
    plt.show()


def plot_explained_var(model: Model, param: Parameters, max_dims: int = 10) -> None:
    """Plot explained variance per dimension.

    Args:
        max_dims: maximum number of dimensions to plot

    Return:
        plot of the explained variance
    """
    # explained_percentage

    explained_percentage: NDArray[Any] = model.explained_var_ratio * 100
    if len(explained_percentage) > max_dims:
        explained_percentage = explained_percentage[:max_dims]

    # plot
    plt.figure(figsize=(12, 6))
    plt.bar(range(len(explained_percentage)), explained_percentage, align="center")
    plt.xticks(
        range(len(explained_percentage)),
        range(1, len(explained_percentage) + 1),
        rotation="horizontal",
    )

    # setting labels and title
    plt.title("Explained variance plot")
    plt.ylabel("Percentage of explained variance")
    plt.xlabel("Dimensions")
    plt.show()



import numpy as np
import pandas as pd
from numpy.testing import assert_allclose
from pandas._testing.asserters import assert_series_equal
from pandas.testing import assert_frame_equal

from saiph.reduction.famd import center, fit, scaler, transform
from saiph.reduction.pca import center as center_pca
from saiph.reduction.pca import fit as fit_pca
from saiph.reduction.pca import scaler as scaler_pca

# mypy: ignore-errors


def test_fit_mix() -> None:
    df = pd.DataFrame(
        {
            "tool": ["toaster", "hammer"],
            "score": ["aa", "ab"],
            "size": [1.0, 4.0],
        }
    )

    result, model, _ = fit(df)

    expected_result = pd.DataFrame(
        {
            "Dim. 1": [-1.73, 1.73],
            "Dim. 2": [0.0, 0.0],
        }
    )
    expected_v = np.array(
        [
            [0.57735, 0.408248, -0.408248, -0.408248, 0.408248],
            [0.816497, -0.288675, 0.288675, 0.288675, -0.288675],
        ]
    )
    expected_s = np.array([1.224745e00, 0.0])
    expected_u = np.array([[-1.0, 1.0], [1.0, 1.0]])
    expected_explained_var = np.array([1.5, 0.0])
    expected_explained_var_ratio = np.array([1.0, 0.0])

    assert_frame_equal(result, expected_result, check_exact=False, atol=0.01)
    assert_frame_equal(model.df, df)
    assert_allclose(model.V, expected_v, atol=0.01)
    assert_allclose(model.s, expected_s, atol=0.01)
    assert_allclose(model.U, expected_u, atol=0.01)
    assert_allclose(model.explained_var, expected_explained_var, atol=0.01)
    assert_allclose(model.explained_var_ratio, expected_explained_var_ratio, atol=0.01),
    assert_allclose(model.variable_coord, model.V.T)
    assert np.array_equal(
        model._modalities, ["tool_hammer", "tool_toaster", "score_aa", "score_ab"]
    )
    # Pertinent ?
    # assert_allclose(model.D_c,np.array([[2.0, 0.0, 0.0],
    # [0.0, 2.0, 0.0], [0.0, 0.0, 1.41421356]]), atol=0.01)
    assert model.D_c is None

    assert_allclose(model.mean, np.array(2.5))
    assert_allclose(model.std, np.array(1.5))

    assert_series_equal(
        model.prop.reset_index(drop=True),
        pd.Series([0.5, 0.5, 0.5, 0.5]).reset_index(drop=True),
        atol=0.01,
    )
    assert np.array_equal(
        model._modalities, ["tool_hammer", "tool_toaster", "score_aa", "score_ab"]
    )


def test_transform() -> None:
    df = pd.DataFrame(
        {
            "tool": ["toaster", "hammer"],
            "score": ["aa", "ab"],
            "size": [1.0, 4.0],
            "age": [55, 62],
        }
    )

    _, model, param = fit(df)

    df_transformed = transform(df, model, param)
    df_expected = pd.DataFrame(
        {
            "Dim. 1": [2.0, -2],
            "Dim. 2": [0.0, 0],
        }
    )

    assert_frame_equal(df_transformed, df_expected)


def test_transform_vs_coord() -> None:
    df = pd.DataFrame(
        {
            "tool": ["toaster", "hammer"],
            "score": ["aa", "ab"],
            "size": [1.0, 4.0],
            "age": [55, 62],
        }
    )

    coord, model, param = fit(df)
    df_transformed = transform(df, model, param)

    assert_frame_equal(df_transformed, coord)


def test_fit_zero() -> None:
    df = pd.DataFrame(
        {
            "tool": ["toaster", "toaster"],
            "score": ["aa", "aa"],
        }
    )

    result, _, _ = fit(df)

    expected = pd.DataFrame(
        {
            "Dim. 1": [0.0, 0.0],
            "Dim. 2": [0.0, 0.0],
        }
    )
    assert_frame_equal(result, expected, check_exact=False, atol=0.01)


def test_scaler_pca_famd() -> None:
    original_df = pd.DataFrame(
        {
            "tool": ["toaster", "hammer"],
            "score": ["aa", "ab"],
            "size": [1.0, 4.0],
            "age": [55, 62],
        }
    )

    _, model, param = fit(original_df)
    df = scaler(model, param)

    _, model_pca, _ = fit_pca(original_df[param.quanti])
    df_pca = scaler_pca(model_pca)

    assert_frame_equal(df[param.quanti], df_pca[param.quanti])


def test_center_pca_famd() -> None:
    original_df = pd.DataFrame(
        {
            "tool": ["toaster", "hammer"],
            "score": ["aa", "ab"],
            "size": [1.0, 4.0],
            "age": [55, 62],
        }
    )

    _, _, param = fit(original_df)
    df, mean1, std1, _, _ = center(original_df, quali=param.quali, quanti=param.quanti)

    df_pca, mean2, std2 = center_pca(original_df[param.quanti])

    assert_frame_equal(df[param.quanti], df_pca[param.quanti])

    assert_series_equal(mean1, mean2)
    assert_series_equal(std1, std2)



"""FAMD projection."""
import sys
from itertools import chain, repeat
from typing import Any, List, Optional, Tuple

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from saiph.models import Model, Parameters
from saiph.reduction.utils.check_params import fit_check_params
from saiph.reduction.utils.common import (
    column_names,
    explain_variance,
    row_weights_uniform,
)
from saiph.reduction.utils.svd import SVD


def fit(
    df: pd.DataFrame,
    nf: Optional[int] = None,
    col_w: Optional[NDArray[Any]] = None,
    scale: Optional[bool] = True,
) -> Tuple[pd.DataFrame, Model, Parameters]:
    """Fit a FAMD model on data.

    Parameters
    ----------
    df: pd.DataFrame
        Data to project.
    nf: int, default: min(df.shape)
        Number of components to keep.
    col_w: np.ndarrayn default: np.ones(df.shape[1])
        Weight assigned to each variable in the projection
        (more weight = more importance in the axes).
    scale: bool
        Unused. Kept for compatibility with model enabling scale=True|False.

    Returns
    -------
    coord: pd.DataFrame
        The transformed data.
    model: Model
        The model for transforming new data.
    param: Parameters
        The parameters for transforming new data.
    """
    nf = nf or min(df.shape)
    _col_weights: NDArray[Any] = col_w or np.ones(df.shape[1])
    if not isinstance(df, pd.DataFrame):
        df = pd.DataFrame(df)
    fit_check_params(nf, _col_weights, df.shape[1])

    # select the categorical and continuous columns
    quanti = df.select_dtypes(include=["int", "float", "number"]).columns.values
    quali = df.select_dtypes(exclude=["int", "float", "number"]).columns.values

    row_w = row_weights_uniform(len(df))
    col_weights = _col_weights_compute(df, _col_weights, quanti, quali)

    df_scale, mean, std, prop, _modalities = center(df, quanti, quali)

    # apply the weights
    Z = ((df_scale * col_weights).T * row_w).T

    # compute the svd
    _U, s, _V = SVD(Z)
    U = ((_U.T) / np.sqrt(row_w)).T
    V = _V / np.sqrt(col_weights)

    explained_var, explained_var_ratio = explain_variance(s, df, nf)

    U = U[:, :nf]
    s = s[:nf]
    V = V[:nf, :]
    columns = column_names(nf)
    coord = df_scale @ V.T
    coord.columns = columns

    model = Model(
        df=df,
        U=U,
        V=V,
        s=s,
        explained_var=explained_var,
        explained_var_ratio=explained_var_ratio,
        variable_coord=pd.DataFrame(V.T),
        mean=mean,
        std=std,
        prop=prop,
        _modalities=_modalities,
        type="famd",
    )

    param = Parameters(
        nf=nf,
        col_w=col_weights,
        row_w=row_w,
        columns=columns,
        quanti=quanti,
        quali=quali,
    )

    return coord, model, param


def _col_weights_compute(
    df: pd.DataFrame, col_w: NDArray[Any], quanti: List[int], quali: List[int]
) -> NDArray[Any]:
    """Calculate weights for columns given what weights the user gave."""
    # Set the columns and row weights
    weight_df = pd.DataFrame([col_w], columns=df.columns)
    weight_quanti = weight_df[quanti]
    weight_quali = weight_df[quali]

    # Get the number of modality for each quali variable
    modality_numbers = []
    for column in weight_quali.columns:
        modality_numbers += [len(df[column].unique())]

    # Set weight vector for categorical columns
    weight_quali_rep = list(
        chain.from_iterable(
            repeat(i, j) for i, j in zip(list(weight_quali.iloc[0]), modality_numbers)
        )
    )

    _col_w: NDArray[Any] = np.array(list(weight_quanti.iloc[0]) + weight_quali_rep)

    return _col_w


def center(
    df: pd.DataFrame, quanti: List[int], quali: List[int]
) -> Tuple[pd.DataFrame, float, float, NDArray[Any], NDArray[Any]]:
    """Center data, scale it and compute modalities.

    Used as internal function during fit. Scaler is better suited when a Model is already fitted.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame to center.
    quanti: np.ndarray
        Indexes of continous variables.
    quali: np.ndarray
        Indexes of categorical variables.

    Returns
    -------
    df_scale: pd.DataFrame
        The scaled DataFrame.
    mean: float
        Mean of the input dataframe.
    std: float
        Standard deviation of the input dataframe. Returns nan as std if no std was asked.
    prop: np.ndarray
        Proportion of each categorical.
    _modalities: np.ndarray
        Modalities for the MCA.
    """
    # Scale the continuous data
    df_quanti = df[quanti]
    mean = np.mean(df_quanti, axis=0)
    df_quanti -= mean
    std = np.std(df_quanti, axis=0)
    std[std <= sys.float_info.min] = 1
    df_quanti /= std

    # scale the categorical data
    df_quali = pd.get_dummies(df[quali].astype("category"))
    prop = np.mean(df_quali, axis=0)
    df_quali -= prop
    df_quali /= np.sqrt(prop)
    _modalities = df_quali.columns.values

    df_scale = pd.concat([df_quanti, df_quali], axis=1)

    return df_scale, mean, std, prop, _modalities


def scaler(
    model: Model, param: Parameters, df: Optional[pd.DataFrame] = None
) -> pd.DataFrame:
    """Scale data using fitted model.

    Parameters
    ----------
    model: Model
        Model computed by fit.
    param: Parameters
        Param computed by fit.
    df: pd.DataFrame
        DataFrame to scale.
        If nothing is specified, takes the DataFrame on which the model was fitted.

    Returns
    -------
    df_scaled: pd.DataFrame
        The scaled DataFrame.
    """
    if df is None:
        df = model.df

    if not isinstance(df, pd.DataFrame):
        df = pd.DataFrame(df)

    df_quanti = df[param.quanti]
    df_quanti = (df_quanti - model.mean) / model.std

    # scale
    df_quali = pd.get_dummies(df[param.quali].astype("category"))
    if model._modalities is not None:
        for mod in model._modalities:
            if mod not in df_quali:
                df_quali[mod] = 0
    df_quali = df_quali[model._modalities]
    df_quali = (df_quali - model.prop) / np.sqrt(model.prop)

    df_scaled = pd.concat([df_quanti, df_quali], axis=1)
    return df_scaled


def transform(df: pd.DataFrame, model: Model, param: Parameters) -> pd.DataFrame:
    """Scale and project into the fitted numerical space.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame to transform.
    model: Model
        Model computed by fit.
    param: Parameters
        Param computed by fit.

    Returns
    -------
    coord: pd.DataFrame
        Coordinates of the dataframe in the fitted space.
    """
    df_scaled = scaler(model, param, df)
    coord = df_scaled @ model.V.T
    coord.columns = param.columns
    return coord


def stats(model: Model, param: Parameters) -> Parameters:
    """Compute contributions and cos2 for each variable."""
    if param.quanti is None or model.U is None or model.s is None:
        raise ValueError(
            "empty param, run fit function to create Model class and Parameters class objects"
        )

    df = pd.DataFrame(scaler(model, param))
    df2: NDArray[Any] = np.array(pd.DataFrame(df).applymap(lambda x: x ** 2))

    # svd of x with row_w and col_w
    weightedTc = _rmultiplication(
        _rmultiplication(df.T, np.sqrt(param.col_w)).T, np.sqrt(param.row_w)
    )
    U, s, V = SVD(weightedTc.T, svd_flip=False)
    ncp0 = min(len(weightedTc.iloc[0]), len(weightedTc), param.nf)
    U = U[:, :ncp0]
    V = V.T[:, :ncp0]
    s = s[:ncp0]
    tmp = V
    V = U
    U = tmp
    mult = np.sign(np.sum(V, axis=0))

    # final V
    mult1 = pd.DataFrame(
        np.array(pd.DataFrame(np.array(_rmultiplication(pd.DataFrame(V.T), mult)))).T
    )
    V = pd.DataFrame()
    for i in range(len(mult1)):
        V[i] = mult1.iloc[i] / np.sqrt(param.col_w[i])
    V = np.array(V).T
    # final U
    mult1 = pd.DataFrame(
        np.array(pd.DataFrame(np.array(_rmultiplication(pd.DataFrame(U.T), mult)))).T
    )
    U = pd.DataFrame()
    for i in range(len(mult1)):
        U[i] = mult1.iloc[i] / np.sqrt(param.row_w[i])
    U = np.array(U).T
    eig: Any = s ** 2
    # end of the svd

    # compute the contribution
    coord_var = np.array(V[0] * s)
    for i in range(1, len(V[:, 0])):
        coord_var = np.vstack((coord_var, V[i] * s))
    contrib_var = (((((coord_var ** 2) / eig).T) * param.col_w).T) * 100
    # compute cos2
    dfrow_w = np.array(pd.DataFrame((df2.T) * param.row_w).T)
    dist2 = []
    for i in range(len(dfrow_w[0])):
        dist2 += [np.sum(dfrow_w[:, i])]
        if abs(abs(dist2[i]) - 1) < 0.001:
            dist2[i] = 1

    cor = ((coord_var.T) / np.sqrt(dist2)).T
    cos2 = cor ** 2

    # compute eta2
    model.df.index = range(len(model.df))
    dfquali = model.df[param.quali]
    eta2 = []
    fi = 0
    coord = pd.DataFrame(
        model.U[:, :ncp0] * model.s[:ncp0], columns=param.columns[:ncp0]
    )
    mods = []
    # for each qualitative column in the original data set
    for count, col in enumerate(dfquali.columns):
        dummy = pd.get_dummies(dfquali[col].astype("category"))
        mods += [len(dummy.columns) - 1]
        # for each dimension
        dim = []
        for j, coordcol in enumerate(coord.columns):
            # for each modality of the qualitative column
            p = 0
            for i in range(len(dummy.columns)):
                p += (
                    np.array(dummy.T)[i] * coord[coordcol] * param.row_w
                ).sum() ** 2 / model.prop[fi + i]
            dim += [p]
        eta1 = (
            np.array(dim) / np.array((coord ** 2).T * param.row_w).sum(axis=1).tolist()
        )
        eta2 += [eta1]
        fi += len(dummy.columns)

        cos2 = cos2[: len(param.quanti)]

    cos2 = cos2 ** 2
    eta2 = np.array(eta2) ** 2
    eta2 = (pd.DataFrame(eta2).T / mods).T

    cos2 = np.concatenate([cos2, eta2], axis=0)  # type: ignore
    param.contrib = contrib_var
    param.cos2 = cos2
    return param


def _rmultiplication(F: pd.DataFrame, marge: NDArray[Any]) -> pd.DataFrame:
    """Multiply each column with the same vector."""
    df_dict = F.to_dict("list")
    for col in df_dict.keys():
        df_dict[col] = df_dict[col] * marge
    df = pd.DataFrame.from_dict(df_dict)
    df.index = F.index
    return df



"""MCA projection."""
from itertools import chain, repeat
from typing import Any, Optional, Tuple

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from saiph.models import Model, Parameters
from saiph.reduction.utils.check_params import fit_check_params

# from scipy.sparse import diags
from saiph.reduction.utils.common import (
    column_names,
    diag,
    explain_variance,
    row_weights_uniform,
)
from saiph.reduction.utils.svd import SVD


def fit(
    df: pd.DataFrame,
    nf: Optional[int] = None,
    col_w: Optional[NDArray[Any]] = None,
    scale: Optional[bool] = True,
) -> Tuple[pd.DataFrame, Model, Parameters]:
    """Fit a MCA model on data.

    Parameters
    ----------
    df: pd.DataFrame
        Data to project.
    nf: int, default: min(df.shape)
        Number of components to keep.
    col_w: np.ndarrayn default: np.ones(df.shape[1])
        Weight assigned to each variable in the projection
        (more weight = more importance in the axes).
    scale: bool
        Unused. Kept for compatibility with model enabling scale=True|False.

    Returns
    -------
    coord: pd.DataFrame
        The transformed data.
    model: Model
        The model for transforming new data.
    param: Parameters
        The parameters for transforming new data.
    """
    nf = nf or min(df.shape)
    _col_weights = col_w or np.ones(df.shape[1])
    if not isinstance(df, pd.DataFrame):
        df = pd.DataFrame(df)
    fit_check_params(nf, _col_weights, df.shape[1])

    # initiate row and columns weights
    row_w = row_weights_uniform(len(df))

    modality_numbers = []
    for column in df.columns:
        modality_numbers += [len(df[column].unique())]
    col_weights: NDArray[Any] = np.array(
        list(
            chain.from_iterable(
                repeat(i, j) for i, j in zip(_col_weights, modality_numbers)
            )
        )
    )

    df_scale, _modalities, r, c = center(df)
    df_scale, T, D_c = _diag_compute(df_scale, r, c)

    # get the array gathering proportion of each modality among individual (N/n)
    df_dummies = pd.get_dummies(df.astype("category"))
    dummies_col_prop = len(df_dummies) / df_dummies.sum(axis=0)

    # apply the weights and compute the svd
    Z = ((T * col_weights).T * row_w).T
    U, s, V = SVD(Z)

    explained_var, explained_var_ratio = explain_variance(s, df, nf)

    U = U[:, :nf]
    s = s[:nf]
    V = V[:nf, :]

    columns = column_names(nf)[: min(df_scale.shape)]
    coord = df_scale @ D_c @ V.T
    coord.columns = columns

    model = Model(
        df=df,
        U=U,
        V=V,
        explained_var=explained_var,
        explained_var_ratio=explained_var_ratio,
        variable_coord=D_c @ V.T,
        _modalities=_modalities,
        D_c=D_c,
        type="mca",
    )

    param = Parameters(
        nf=nf,
        col_w=col_weights,
        row_w=row_w,
        columns=columns,
        dummies_col_prop=dummies_col_prop,
    )

    return coord, model, param


def center(
    df: pd.DataFrame,
) -> Tuple[pd.DataFrame, NDArray[Any], NDArray[Any], NDArray[Any]]:
    """Center data and compute modalities.

    Used as internal function during fit. Scaler is better suited when a Model is already fitted.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame to center.

    Returns
    -------
    df_centered: pd.DataFrame
        The centered DataFrame.
    _modalities: np.ndarray
        Modalities for the MCA
    r: np.ndarray
        Sums line by line
    c: np.ndarray
        Sums column by column
    """
    df_scale = pd.get_dummies(df.astype("category"))
    _modalities = df_scale.columns.values

    # scale data
    df_scale /= df_scale.sum().sum()

    c = np.sum(df_scale, axis=0)
    r = np.sum(df_scale, axis=1)
    return df_scale, _modalities, r, c


def scaler(model: Model, df: Optional[pd.DataFrame] = None) -> pd.DataFrame:
    """Scale data using fitted model.

    Parameters
    ----------
    model: Model
        Model computed by fit.
    df: pd.DataFrame
        DataFrame to scale.
        If nothing is specified, takes the DataFrame on which the model was fitted.

    Returns
    -------
    df_scaled: pd.DataFrame
        The scaled DataFrame.
    """
    if df is None:
        df = model.df

    if not isinstance(df, pd.DataFrame):
        df = pd.DataFrame(df)

    df_scaled = pd.get_dummies(df.astype("category"))
    if model._modalities is not None:
        for mod in model._modalities:
            if mod not in df_scaled:
                df_scaled[mod] = 0
    df_scaled = df_scaled[model._modalities]

    # scale
    df_scaled /= df_scaled.sum().sum()
    df_scaled /= np.array(np.sum(df_scaled, axis=1))[:, None]
    return df_scaled


def _diag_compute(
    df_scale: pd.DataFrame, r: NDArray[Any], c: NDArray[Any]
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Compute diagonal matrices and scale data."""
    eps = np.finfo(float).eps
    if df_scale.shape[0] >= 10000:
        D_r = diag(1 / (eps + np.sqrt(r)), use_scipy=True)
    else:
        D_r = diag(1 / (eps + np.sqrt(r)), use_scipy=False)
    D_c = diag(1 / (eps + np.sqrt(c)), use_scipy=False)

    T = D_r @ (df_scale - np.outer(r, c)) @ D_c
    return df_scale / np.array(r)[:, None], T, D_c


def transform(df: pd.DataFrame, model: Model, param: Parameters) -> pd.DataFrame:
    """Scale and project into the fitted numerical space.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame to transform.
    model: Model
        Model computed by fit.
    param: Parameters
        Param computed by fit.

    Returns
    -------
    coord: pd.DataFrame
        Coordinates of the dataframe in the fitted space.
    """
    df_scaled = scaler(model, df)
    coord = df_scaled @ model.D_c @ model.V.T
    coord.columns = param.columns
    return coord


def stats(model: Model, param: Parameters) -> Parameters:
    """Compute the contributions of each variable in each axis."""
    V = np.dot(model.D_c, model.V.T)  # type: ignore
    total = pd.get_dummies(model.df.astype("category")).sum().sum()
    df = pd.get_dummies(model.df.astype("category"))
    F = df / total

    # Column and row weights
    marge_col = F.sum(axis=0)
    marge_row = F.sum(axis=1)
    fsurmargerow = _rdivision(F, marge_row)
    fmargerowT = pd.DataFrame(
        np.array(fsurmargerow).T,
        columns=list(fsurmargerow.index),
        index=list(fsurmargerow.columns),
    )
    fmargecol = _rdivision(fmargerowT, marge_col)
    Tc = (
        pd.DataFrame(
            np.array(fmargecol).T,
            columns=list(fmargecol.index),
            index=list(fmargecol.columns),
        )
        - 1
    )

    # Weights and svd of Tc
    weightedTc = _rmultiplication(
        _rmultiplication(Tc.T, np.sqrt(marge_col)).T, np.sqrt(marge_row)
    )
    U, s, V = SVD(weightedTc.T, svd_flip=False)
    ncp0 = min(len(weightedTc.iloc[0]), len(weightedTc), param.nf)
    U = U[:, :ncp0]
    V = V.T[:, :ncp0]
    s = s[:ncp0]
    tmp = V
    V = U
    U = tmp
    mult = np.sign(np.sum(V, axis=0))

    # final V
    mult1 = pd.DataFrame(
        np.array(pd.DataFrame(np.array(_rmultiplication(pd.DataFrame(V.T), mult)))).T
    )
    V = pd.DataFrame()
    for i in range(len(mult1)):
        V[i] = mult1.iloc[i] / np.sqrt(marge_col[i])
    V = np.array(V).T

    # final U
    mult1 = pd.DataFrame(
        np.array(pd.DataFrame(np.array(_rmultiplication(pd.DataFrame(U.T), mult)))).T
    )
    U = pd.DataFrame()
    for i in range(len(mult1)):
        U[i] = mult1.iloc[i] / np.sqrt(marge_row[i])
    U = np.array(U).T

    # computing the contribution
    eig: Any = s ** 2

    for i in range(len(V[0])):
        V[:, i] = V[:, i] * np.sqrt(eig[i])
    coord_col = V

    for i in range(len(U[0])):
        U[:, i] = U[:, i] * np.sqrt(eig[i])

    coord_col = coord_col ** 2

    for i in range(len(coord_col[0])):
        coord_col[:, i] = (coord_col[:, i] * marge_col) / eig[i]

    param.contrib = coord_col * 100

    return param


def _rmultiplication(F: pd.DataFrame, marge: NDArray[Any]) -> pd.DataFrame:
    """Multiply each column with the same vector."""
    df_dict = F.to_dict("list")
    for col in df_dict.keys():
        df_dict[col] = df_dict[col] * marge
    df = pd.DataFrame.from_dict(df_dict)
    df.index = F.index
    return df


def _rdivision(F: pd.DataFrame, marge: NDArray[Any]) -> pd.DataFrame:
    """Divide each column with the same vector."""
    df_dict = F.to_dict("list")
    for col in df_dict.keys():
        df_dict[col] = df_dict[col] / marge
    df = pd.DataFrame.from_dict(df_dict)
    df.index = F.index
    return df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['saiph', 'saiph.reduction', 'saiph.reduction.utils']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.4,<4.0',
 'numpy>=1.21,<2.0',
 'pandas>=1.3,<2.0',
 'scikit-learn>=1.0,<2.0',
 'scipy>=1.7,<2.0']

setup_kwargs = {
    'name': 'saiph',
    'version': '0.2.1',
    'description': 'A projection package',
    'long_description': None,
    'author': 'Zineb Bennis',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



from typing import Any, Tuple

import pandas as pd
from numpy.typing import NDArray
from scipy import linalg
from sklearn.utils import extmath


def SVD(
    df: pd.DataFrame, svd_flip: bool = True
) -> Tuple[NDArray[Any], NDArray[Any], NDArray[Any]]:
    """Compute Singular Value Decomposition.

    Parameters
    ----------
    df: pd.DataFrame
        Matrix to decompose.
    svd_flip: bool
        Whether to use svd_flip on U and V or not.

    Returns
    -------
    U: np.ndarray
        Unitary matrix having left singular vectors as columns.
    s: np.ndarray
        The singular values.
    V: np.ndarray
        Unitary matrix having right singular vectors as rows.
    """
    U, s, V = linalg.svd(df, full_matrices=False)
    if svd_flip:
        U, V = extmath.svd_flip(U, V)
    return U, s, V



from .projection import fit, inverse_transform, stats, transform
from .visualization import plot_circle, plot_explained_var, plot_var_contribution

# Also modify in pyproject.toml
__version__ = "0.2.1"

__all__ = [
    "__version__",
    "fit",
    "inverse_transform",
    "transform",
    "stats",
    "plot_circle",
    "plot_explained_var",
    "plot_var_contribution",
]
